<?php
/**
 * Carerix Wordpress Plugin
 *
 * @category template manager
 * @author Bogdan Ivanov
 * @copyright Copyright (c) 2013 Carerix.com (http://www.carerix.com)
 * @link http://www.carerix.com
 * @since Aug 26, 2013 11:36:11 AM
 * @version $Rev$
 */

/**
 * Settings class
 * Used to ....
 */
class Admin_Forms extends WP_Page {

    protected $interfacer = null;

    public function __construct() {
        $this->interfacer = WP_Interfacer::getInstance();
    }

    public function getCapability() {
        return WP_Constants::CAPABILITY_ADMIN;
    }

    public function getMenuSlug() {
        return 'carerix_forms_page';
    }

    public function getMenuTitle() {
        return 'Forms';
    }

    public function getPageTitle() {
        return 'Forms';
    }

    public function page() {
        $input_data = filter_input_array(INPUT_GET, FILTER_SANITIZE_STRING);
        switch ($input_data['action']) {
            case 'edit':
                $this->formScreen($input_data['id']);
                break;
            case 'add':
                $this->formScreen();
                break;
            case 'delete':
                $this->deleteForm($input_data['id']);
            case 'list':
            default:
                $this->listForms();
                break;
        }
    }

    private function listForms() {
        $data = array();

        $formsTable = new WP_Table(Environment_Installation::TABLE_FORMS);
        $applyForms = $formsTable->getByType(WP_Constants::ENTITY_TYPE_FORM_APPLY);

        foreach ($applyForms as &$applyForm) {
            $applyForm['settings'] = unserialize($applyForm['settings']);
            $applyForm['editLink'] = $this->getSourceEditLink($applyForm['id']);
            $applyForm['deleteLink'] = $this->getSourceDeleteLink($applyForm['id']);
        }

        $data['add_new'] = $this->createActionLink('add');
        $data['applyForms'] = $applyForms;

        $template = new Utils_Template('admin/forms/apply_list');
        $template->with($data)->display();
    }

    private function formScreen($id = false) {
        $data = array();

        $args = array(
            'sort_order' => 'ASC',
            'sort_column' => 'post_title',
            'hierarchical' => 1,
            'exclude' => '',
            'include' => '',
            'meta_key' => '',
            'meta_value' => '',
            'authors' => '',
            'child_of' => 0,
            'parent' => -1,
            'exclude_tree' => '',
            'number' => '',
            'offset' => 0,
            'post_type' => 'page',
            'post_status' => 'publish'
        );
        $_allpages = get_pages($args);
        $allpages = array();
        foreach ($_allpages as $p) {
            $allpages[$p->ID] = $p->guid;
        }
        $data['pageurls'] = $allpages;

        $labels = Utils_Labels::get($this->interfacer->language_major_current());
        $fieldsToDisplay = array();

        if (Environment_Settings::getFormAddressFormatGeneral() == 'international') {
            Entity_Form_EmployeeApply::$fields['address']['items'] = Entity_Form_EmployeeApply::addressInternationalFormat();
        }
        $fieldsAvailable = Entity_Form_EmployeeApply::$fields;
        $input_data = filter_input_array(INPUT_POST, FILTER_SANITIZE_STRING);
        $availability = array();
        if ($input_data != NULL && wp_verify_nonce($input_data['_carerix_nonce_field'], 'apply')) {

            if (isset($input_data['post_title']) && trim($input_data['post_title']) != '') {
                $id = false;
                if (isset($input_data['id']))
                    $id = $input_data['id'];
                if (isset($input_data['carerix_sync'])) {
                    Utils_Cache::cleanEnvironment();
                    try {
                        $availability = $this->getSystemFieldsAvailability();
                    } catch (Exception $ex) {
                        $data['sync_error'] = true;
                    }
                    if (!isset($data['sync_error'])) {
                        $data['sync_success'] = true;
                    }
                } else {
                    unset($input_data['save']);

                    $backup_data = $input_data;

                    $applyForm = new Entity_Form_Apply($id);
                    $applyForm->save($input_data);
                    $id = $applyForm->getID();

                    $data['form_success'] = true;
                }
            } else {
                $data['page_error'] = true;
                $data['title_error'] = true;
            }
        }

        if ($id) {
            $applyForm = new Entity_Form_Apply($id);

            $formData = $applyForm->getData();
            $formData['settings'] = unserialize($formData['settings']);
            $data['edit_data'] = $formData;

            if (isset($formData['settings']['field_order'])) {
                $field_order = json_decode(html_entity_decode($formData['settings']['field_order']));
                if ($field_order) {
                    $_collector = array();
                    foreach ($field_order as $section => $fields) {
                        if (!isset($_collector[$section]))
                            $_collector[$section] = array();

                        $_collector[$section] = $fieldsAvailable[$section];
                        $_collector[$section]['items'] = array();

                        if ($section == 'address') {
                            if (Environment_Settings::getFormAddressFormatGeneral() == 'international') {
                                $addressFields = Entity_Form_EmployeeApply::addressInternationalFormat();
                                $fields = array_keys($addressFields);
                            } else {

                                $dutch_fields = array_keys(Entity_Form_EmployeeApply::$fields[$section]['items']);

                                $fields = $dutch_fields;
                            }
                        }

                        foreach ($fields as $field) {
                            $_collector[$section]['items'][$field] = $fieldsAvailable[$section]['items'][$field];
                        }

                        foreach ($fieldsAvailable as $field_name => $field_cfg) {
                            if (!isset($_collector[$section]['items'][$field_name])) {
                                $_collector[$section]['items'][$field_name] = $field_cfg;
                            }
                        }
                    }
                    $fieldsAvailable = $_collector;
                }
            }
        }
        $data['available'] = $availability;
        $data['cancel_link'] = $this->createActionLink('list');

        foreach ($fieldsAvailable as $section => $config) {
            $fields = $config['items'];
            foreach ($fields as $field_name => $field_config) {
                $fieldsToDisplay[$field_name] = array(
                    'section_type' => @$config['type'],
                    'field_type' => @$field_config['type'],
                    'section_subtype' => @$field_config['subtype'],
                    'section' => $section,
                    'section_label' => @$labels[$config['label']],
                    'label' => @$labels[$field_config['label']]);
            }
        }

        $system_countries = Entity_Form_EmployeeApply::get_system_select_options('Country');
        $countries = array();

        foreach ($system_countries as $scountry) {
            $countries[$scountry['value']] = $scountry['label'];
        }

        $data['defaults']['fields'] = array(
            'country' => $countries);



        $data['defaults']['selected'] = array(
            'country' => @$formData['settings']['default_country']
        );
        $data['fieldsToDisplay'] = $fieldsToDisplay;

        $template = new Utils_Template('admin/forms/apply_edit');
        $template->with($data)->display();
    }

    private function deleteForm($id) {
        $formsTable = new WP_Table(Environment_Installation::TABLE_FORMS);
        $formsTable->delete($id);
    }

    private function createActionLink($action_name) {
        return 'admin.php?page=' . $this->getParentTab()->getMenuSlug() .
                '&tab=' . $this->getMenuSlug() .
                '&action=' . $action_name;
    }

    private function getSourceEditLink($source_id) {
        return $this->createActionLink('edit') . '&id=' . $source_id;
    }

    private function getSourceDeleteLink($source_id) {
        return $this->createActionLink('delete') . '&id=' . $source_id;
    }

    private function getSystemFieldsAvailability() {
        return Entity_Form_Apply::systemAvailability();
    }

}
